use std::collections::VecDeque;
use std::env;
use std::fs::File;
use std::io;
use std::io::BufRead;
use std::io::{Read, Write};
use std::mem;
use std::process;

/// The standard program exit codes that could be used when this
/// program terminates.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum ExitReason {
    Success = 0,
    Usage = 64,
    Syntax = 65,
    OpenInput = 66,
    Runtime = 70,
    Io = 74,
}
impl ExitReason {
    /// Exits the program with this reason.
    fn exit(self) -> ! {
        process::exit(self as i32);
    }
}

/// A trait for things that can exit the program upon encountering an error.
trait ExitOnError {
    /// The type values contain when they are not errors.
    type Success;

    /// If this is an error, exit the program.
    ///
    /// If this is not an error, return the success type.
    fn exit_on_error(self) -> Self::Success;
}
impl<T> ExitOnError for io::Result<T> {
    type Success = T;

    fn exit_on_error(self) -> T {
        match self {
            Ok(t) => t,
            Err(e) => {
                eprintln!("esimpl: error: could not read input: {e}");
                ExitReason::Io.exit();
            }
        }
    }
}


/// The type used for stanza numbers and semideque elements.
///
/// Stanza numbers and semideque elements are added in order to
/// produce stanza numbers, so they conceptually belong to the same
/// type. A `Stanza` can thus either be a stanza number itself, or
/// else an offset between stanza numbers.
#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
struct Stanza(usize);

/// The type used for semideque numbers.
#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
struct Semideque(usize);

/// Commands that just add data to semideques.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum DataCommandType {
    Push(Semideque),
    PushBack(Semideque),
    Output,
}

#[derive(Debug, Clone)]
struct DataCommand {
    cmd_type: DataCommandType,
    pushed_values: Box<[Stanza]>,
}

/// Commands that do control flow.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum ControlCommand {
    Halt,
    Goto { stanza: Stanza, linked_to: Semideque },
    InputGoto(Stanza),
    PopGoto { pop_from: Semideque, add: Stanza },
}

/// The types of table link.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum TableLink {
    SemidequeLinked(Semideque),
    IoLinked
}

/// The commands that make up a single stanza.
#[derive(Debug, Clone)]
struct StanzaCommands {
    data_commands: Box<[DataCommand]>,
    control_command: ControlCommand,
    table_linked_to: Option<TableLink>,
    new_table: bool,
}

#[derive(Debug, Clone)]
struct Program(Box<[StanzaCommands]>);

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum EsiSyntax {
    Binary,
    Text
}

/// Parses the given stream into a program in the given syntax.
///
/// This function will print errors to standard error and exit the
/// program, rather than returning them.
///
/// The program is read from the specified buffered reader; anything
/// beyond the end of the program will be left in its buffer.
fn parse_program<T: io::Read>(br: &mut io::BufReader<T>, syntax: EsiSyntax) -> Program {
    let mut stanzas = Vec::new();

    // A stanza builder contains the partial information known about a stanza that's
    // being parsed: the data commands that have been seen so far, and its table link
    // information.
    #[derive(Default)]
    struct StanzaBuilder {
        table_linked_to: Option<TableLink>,
        current_data_commands: Vec<DataCommand>,
        new_table: bool,
    }
    let mut sb = StanzaBuilder::default();

    fn push_stanza(stanzas: &mut Vec<StanzaCommands>,
                   control_command: ControlCommand,
                   sb: &mut StanzaBuilder) {
        stanzas.push(StanzaCommands {
            data_commands: mem::take(&mut sb.current_data_commands).into_boxed_slice(),
            control_command,
            table_linked_to: sb.table_linked_to,
            new_table: mem::take(&mut sb.new_table)
        });
    }
    fn push_data_command(cmd_type: DataCommandType,
                         pushed_values: Box<[Stanza]>,
                         sb: &mut StanzaBuilder) {
        sb.current_data_commands.push(DataCommand { cmd_type, pushed_values })
    }
    fn start_new_table(sb: &mut StanzaBuilder, linked_to: TableLink, loc: usize, locunits: &str) {
        if !sb.current_data_commands.is_empty() {
            eprintln!("esimpl: {locunits} {loc}: error: table starts in the middle of a stanza");
            ExitReason::Syntax.exit();
        }
        sb.new_table = true;
        sb.table_linked_to = Some(linked_to);
    }
    /// Removes the first datum from the "push" instruction for the given semideque,
    /// and returns it.
    ///
    /// Returns `None` if the instruction is absent or empty.
    fn steal_data_from_semideque(sb: &mut StanzaBuilder, semideque: Semideque) -> Option<Stanza> {
        let pvref: &mut Box<[Stanza]> = &mut sb.current_data_commands.iter_mut().find(
            |dc| dc.cmd_type == DataCommandType::Push(semideque))?.pushed_values;
        if pvref.is_empty() { return None; }

        // Swap a placeholder over the push instruction so that it can be
        // reallocated in a panic-safe way. (Not that the panic-safety
        // really matters, but the compiler will notice if it isn't panic-
        // safe and refuse to compile the program.)
        let mut pv = mem::replace(pvref, Vec::new().into_boxed_slice()).into_vec();
        let rv = pv.remove(0);
        *pvref = pv.into_boxed_slice();
        Some(rv)
    }

    if syntax == EsiSyntax::Text {
        fn assert_empty(i: &mut impl Iterator, cmd: &str, lineno: usize) {
            if i.next().is_some() {
                eprintln!("esimpl: line {lineno}: error: extra text after {cmd} command");
                ExitReason::Syntax.exit();
            }
        }
        fn parse_int(w: &str, what: &str, lineno: usize) -> usize {
            match w.parse() {
                Err(e) => {
                    eprintln!("esimpl: line {lineno}: error: expected {what}, saw {w:?} ({e})");
                    ExitReason::Syntax.exit();
                },
                Ok(rv) => rv
            }
        }
        fn parse_int_take<'a, T>(i: &mut T, what: &str, lineno: usize) -> usize
        where T: Iterator<Item = &'a str> {
            let Some(w) = i.next() else {
                eprintln!("esimpl: line {lineno}: error: expected {what}, saw end of line");
                ExitReason::Syntax.exit();
            };
            parse_int(w, what, lineno)
        }
        fn parse_pushed_values<'a, T>(i: &mut T, cmd: &str, lineno: usize) -> Box<[Stanza]>
        where T: Iterator<Item = &'a str> {
            i.map(|w| match w.parse() {
                Err(e) => {
                    eprintln!("esimpl: line {lineno}: error: \
                               expected data for {cmd} command, saw {w:?} ({e})");
                    ExitReason::Syntax.exit();
                },
                Ok(rv) => Stanza(rv)
            }).collect::<Vec<_>>().into_boxed_slice()
        }

        let mut lineno = 0;
        for line in br.lines() {
            lineno += 1;
            let line = line.exit_on_error();

            // strip comments
            let line = line.split('#').next()
                .expect("str.split() should return at least 1 component");

            // the only valid whitespace is ASCII space
            let line = line.trim_matches(' ');

            // there are seven valid types of line – we use the control flow
            // here as a state machine to handle them
            let mut words = line.split(' ').filter(|w| !w.is_empty());
            match words.next() {
                // a blank line, or comment-only line
                None => (),

                // a halt instruction
                Some("h" | "halt") => {
                    assert_empty(&mut words, "halt", lineno);
                    push_stanza(&mut stanzas, ControlCommand::Halt, &mut sb);
                },

                // an input-goto instruction
                Some("i" | "input" | "input-goto") => {
                    let table = parse_int_take(&mut words, "input-goto table", lineno);
                    assert_empty(&mut words, "input-goto", lineno);
                    push_stanza(&mut stanzas, ControlCommand::InputGoto(Stanza(table)), &mut sb);
                },

                // an output instruction
                Some("o" | "output") =>
                    push_data_command(DataCommandType::Output,
                                      parse_pushed_values(&mut words, "output", lineno), &mut sb),

                // instructions that should have been preceded by a semideque number
                Some("t" | "table" | "p" | "push" | "q" | "pushback" | "j" | "pop-goto" | "g" | "goto") => {
                    eprintln!("esimpl: error: line {lineno}: \
                               this command should be preceded by a semideque number");
                    ExitReason::Syntax.exit();
                },

                // start of a new iotable
                Some("u" | "iotable") =>
                    start_new_table(&mut sb, TableLink::IoLinked, lineno, "line"),

                Some(w) => {
                    let semideque = Semideque(parse_int(w, "semideque number or command", lineno));
                    match words.next() {
                        Some("p" | "push") =>
                            push_data_command(DataCommandType::Push(semideque),
                                              parse_pushed_values(&mut words, "push", lineno), &mut sb),
                        Some("q" | "pushback") =>
                            push_data_command(DataCommandType::PushBack(semideque),
                                              parse_pushed_values(&mut words, "pushback", lineno),
                                              &mut sb),
                        Some("j" | "pop-goto") => {
                            let table = parse_int_take(&mut words, "pop-goto table", lineno);
                            assert_empty(&mut words, "pop-goto", lineno);
                            push_stanza(&mut stanzas, ControlCommand::PopGoto {
                                pop_from: semideque,
                                add: Stanza(table)
                            }, &mut sb);
                        },
                        Some("u" | "iotable" | "h" | "halt" | "i" | "input" | "input-goto" | "o" | "output") => {
                            eprintln!("esimpl: error: line {lineno}: \
                                       this command does not take a semideque number");
                            ExitReason::Syntax.exit();
                        },
                        Some("g" | "goto") => {
                            let target = parse_int_take(&mut words, "goto target", lineno);
                            assert_empty(&mut words, "goto", lineno);
                            push_stanza(&mut stanzas, ControlCommand::Goto {
                                linked_to: semideque,
                                stanza: Stanza(target)
                            }, &mut sb);
                        },
                        Some("t" | "table") => {
                            start_new_table(&mut sb, TableLink::SemidequeLinked(semideque),
                                            lineno, "line");
                        },
                        Some(c) => {
                            eprintln!("esimpl: error: line {lineno}: \
                                       unrecognised command {c:?}");
                            ExitReason::Syntax.exit();
                        },
                        None => {
                            eprintln!("esimpl: error: line {lineno}: \
                                       a number on its own is not a valid command");
                            ExitReason::Syntax.exit();
                        },
                    }
                }
            }
        }
    } else {
        let mut br = br.bytes();
        let mut data_line = Vec::new();

        /// Parses a run of zeroes, then returns the subsequent byte.
        ///
        /// The subsequent byte can't be 0, so 0 is used as a sentinel for EOF.
        fn parse_unary<T: Read>(i: &mut io::Bytes<T>, what: &str, loc: usize) -> (usize, u8) {
            let mut rv: usize = 0;
            loop {
                match i.next().transpose().exit_on_error() {
                    None => break (rv, 0),
                    Some(0) => {
                        let Some(incrv) = rv.checked_add(1) else {
                            // this one probably isn't possible on 64-bit systems, but
                            // could happen on 32-bit systems if given a large enough
                            // input file
                            eprintln!("esimpl: error: stanza {loc}: \
                                       expected {what}, saw overflowing unary constant");
                            ExitReason::Syntax.exit();
                        };
                        rv = incrv;
                    },
                    Some(b) => break (rv, b)
                }
            }
        }

        // stanza 0's data has an unusual encoding, so it gets parsed here first
        let mut ending_byte;
        loop {
            let (data, eb) = parse_unary(&mut br, "stanza 0 data", 0);
            if data != 0 && eb != 1 {
                // This syntax is normally used for a pop-goto command, but those aren't
                // allowed in stanza 0
                eprintln!("esimpl: error: stanza 0: bad unary encoding, 0x00 was \
                           followed by {eb:#02x}");
                ExitReason::Syntax.exit();
            }
            match eb {
                0 => {
                    eprintln!("esimpl: error: stanza 0: unexpected end of file");
                    ExitReason::Syntax.exit();
                },
                1 => data_line.push(Stanza(data)),
                2 => push_data_command(DataCommandType::Push(Semideque(sb.current_data_commands.len())),
                                       mem::take(&mut data_line).into_boxed_slice(), &mut sb),
                // special-case encoding for stanza 0: the goto command starts with 0xD, not 9
                0xD => { ending_byte = 9; break; }
                b => {
                    eprintln!("esimpl: error: stanza 0: expected stanza 0 definition, \
                               saw undefined byte {b:#02x}");
                    ExitReason::Syntax.exit();
                }
            }
        }
        if !data_line.is_empty() {
                    eprintln!("esimpl: error: stanza 0: saw control command before \
                               the terminating 0x02 of the last semideque");
                    ExitReason::Syntax.exit();
        }

        let semideque_count = Semideque(sb.current_data_commands.len());

        // The pop-goto location can be specified twice, but doesn't have to be specified the
        // first time if the amount added is 0. So there are two `pop_goto_where` variables,
        // which are compared to ensure they match.
        let mut pop_goto_where_1 = None;
        let mut pop_goto_add = 0;

        'stanzas: loop {
            // decode the control instruction:
            // - goto/pop-goto consists of 9, then some number of repeats of [3, 2], then 8
            // - input-goto is B, taking the argument implicitly from semideque 0
            // - halt is C
            match ending_byte {
                9 => {
                    // The second pop-goto location always has to be specified.
                    let mut pop_goto_where_2 = 0_i128;
                    let mut next_byte = 3;
                    loop {
                        match br.next().transpose().exit_on_error() {
                            Some(b @ (2 | 3)) if b == next_byte => {
                                if b == 3 { pop_goto_where_2 += 1; }
                                next_byte = next_byte ^ 2 ^ 3;
                            },
                            Some(8) if next_byte == 3 => break,
                            Some(b) => {
                                eprintln!("esimpl: error: stanza {}: unexpected {b:#02x} \
                                           in encoding of goto / pop-goto instruction",
                                          stanzas.len());
                                ExitReason::Syntax.exit();
                            },
                            None => {
                                eprintln!("esimpl: error: stanza {}: end of file in the \
                                           middle of a pop-goto instruction", stanzas.len());
                                ExitReason::Syntax.exit();
                            },
                        }
                    }
                    // If this is a pop-goto, check that the two pop_goto_wheres match
                    // It might also be a regular goto, in which case pop_goto_where_1
                    // will be None
                    if let Some(pgw1) = pop_goto_where_1 {
                        if (pgw1 as i128) != pop_goto_where_2 {
                            eprintln!("esimpl: error: stanza {}: the data of this stanza \
                                       sets up a pop-goto for semideque {pgw1}, but the \
                                       pop-goto command actually pops semideque \
                                       {pop_goto_where_2}", stanzas.len());
                            ExitReason::Syntax.exit();
                        }
                    }
                    if (semideque_count.0 as i128) <= pop_goto_where_2 {
                        eprintln!("esimpl: error: stanza {}: attempt to goto/pop-goto \
                                   semideque {pop_goto_where_2}, but only \
                                   {} semideques exist",
                                  stanzas.len(), semideque_count.0);
                        ExitReason::Syntax.exit();
                    }
                    let semideque = Semideque(pop_goto_where_2.try_into().expect(
                        "semideque number was less than semideque_count, so should be in range"));
                    if pop_goto_where_1.is_some() {
                        push_stanza(&mut stanzas, ControlCommand::PopGoto {
                            pop_from: semideque,
                            add: Stanza(pop_goto_add)
                        }, &mut sb);
                    } else {
                        let Some(stanza) = steal_data_from_semideque(&mut sb, semideque)
                        else {
                            eprintln!("esimpl: error: stanza {}: goto command does not \
                                       specify a goto target", stanzas.len());
                            ExitReason::Syntax.exit();
                        };
                        push_stanza(&mut stanzas, ControlCommand::Goto {
                            stanza, linked_to: semideque }, &mut sb);
                    }
                },
                0xB => {
                    if pop_goto_where_1.is_some() {
                        eprintln!("esimpl: error: stanza {}: a semideque's encoding \
                                   specifies that it uses a pop-goto instruction, but \
                                   the stanza actually has an input instruction", stanzas.len());
                        ExitReason::Syntax.exit();
                    }
                    let Some(table) = steal_data_from_semideque(&mut sb, Semideque(0))
                    else {
                        eprintln!("esimpl: error: stanza {}: input command does not specify \
                                   where to go after reading input", stanzas.len());
                        ExitReason::Syntax.exit();
                    };
                    push_stanza(&mut stanzas, ControlCommand::InputGoto(table), &mut sb);
                },
                0xC => {
                    if pop_goto_where_1.is_some() {
                        eprintln!("esimpl: error: stanza {}: a semideque's encoding \
                                   specifies that it uses a pop-goto instruction, but \
                                   the stanza actually has an input instruction", stanzas.len());
                        ExitReason::Syntax.exit();
                    }
                    push_stanza(&mut stanzas, ControlCommand::Halt, &mut sb);
                },
                0 => {
                    eprintln!("esimpl: error: stanza {}: unexpected end of file, \
                               expected control command", stanzas.len());
                    ExitReason::Syntax.exit();
                },
                b => {
                    eprintln!("esimpl: error: stanza {}: expected control command, \
                               saw invalid byte {b:#02x}", stanzas.len());
                    ExitReason::Syntax.exit();
                },
            }

            pop_goto_where_1 = None;
            pop_goto_add = 0;

            // Is there another stanza to parse?
            let mut table_link_index = 0;
            let mut stanza_header_bytes_remaining = semideque_count.0;
            let mut start_of_stanza_header = true;
            let mut new_table = false;
            while stanza_header_bytes_remaining > 0 {
                match br.next().transpose().exit_on_error() {
                    Some(0xA) => {
                        if !start_of_stanza_header {
                            eprintln!("esimpl: error: stanza {}: unexpected table separator \
                                       seen in the middle of the stanza header (should be \
                                       before it starts)", stanzas.len());
                            ExitReason::Syntax.exit();
                        }
                        new_table = true;
                        start_of_stanza_header = false;
                        continue; // doesn't consume a byte of the header
                    }
                    Some(5) => table_link_index += 1,
                    Some(4) => (),
                    None => {
                        eprintln!("esimpl: error: stanza {}: unexpected end of file, \
                                   expected stanza header", stanzas.len());
                        ExitReason::Syntax.exit();
                    },
                    Some(0xE) if start_of_stanza_header => break 'stanzas,
                    Some(b) => {
                        eprintln!("esimpl: error: stanza {}: expected stanza header, \
                                   saw invalid byte {b:#02x}", stanzas.len());
                        ExitReason::Syntax.exit();
                    }
                }
                start_of_stanza_header = false;
                stanza_header_bytes_remaining -= 1;
            }

            let link = if table_link_index == semideque_count.0 {
                TableLink::IoLinked
            } else {
                TableLink::SemidequeLinked(Semideque(table_link_index))
            };

            if new_table {
                start_new_table(&mut sb, link, stanzas.len(), "stanza");
            } else if Some(link) != sb.table_linked_to {
                eprintln!("esimpl: error: stanza {}: the information about the table \
                           this stanza is linked to does not match the stanza before, \
                           but it is not the first stanza in its table", stanzas.len());
                ExitReason::Syntax.exit();
            }

            let (mut data, mut eb) = parse_unary(&mut br, "stanza data", stanzas.len());
            if eb == 0 {
                eprintln!("esimpl: error: stanza {}: unexpected end of file, \
                           expected stanza data", stanzas.len());
                ExitReason::Syntax.exit();
            }

            let mut semideque = 0;
            let mut seen_push = false;
            loop {
                if data != 0 && eb != 1 {
                    if eb == 3 && data_line.is_empty() {
                        // This is the only valid case for trailing 0s that aren't
                        // followed by a 1: they specify the pop-goto offset for
                        // semideques that weren't pushed onto. (If a stanza is the
                        // target of both a push and a pop-goto, that's actually just
                        // a goto, and the encoding is interpreted as such.)
                        if let Some(sdq) = pop_goto_where_1 {
                            eprintln!("esimpl: error: stanza {}: at least two semideques \
                                       ({sdq} and {semideque}) contain pop-goto add counts",
                                      stanzas.len());
                            ExitReason::Syntax.exit();
                        }
                        pop_goto_where_1 = Some(semideque);
                        pop_goto_add = data;
                    } else {
                        eprintln!("esimpl: error: stanza {}: invalid unary encoding
                                   (zeroes are followed by {eb:#02x})", stanzas.len());
                        ExitReason::Syntax.exit();
                    }
                }
                match eb {
                    1 => data_line.push(Stanza(data)),
                    3 => {
                        if seen_push {
                            eprintln!("esimpl: error: stanza {}: two push instructions \
                                       for semideque {semideque}", stanzas.len());
                            ExitReason::Syntax.exit();
                        }
                        push_data_command(DataCommandType::Push(Semideque(semideque)),
                                          mem::take(&mut data_line).into_boxed_slice(), &mut sb);
                        seen_push = true;
                    },
                    2 => {
                        if !seen_push {
                            eprintln!("esimpl: error: stanza {}: pushback instruction \
                                       for semideque {semideque} with no preceding push",
                                      stanzas.len());
                            ExitReason::Syntax.exit();
                        }
                        push_data_command(DataCommandType::PushBack(Semideque(semideque)),
                                          mem::take(&mut data_line).into_boxed_slice(), &mut sb);
                        seen_push = false;
                        semideque += 1;
                    },
                    0 => {
                        eprintln!("esimpl: error: stanza {}: unexpected end of file, \
                                   expected data command", stanzas.len());
                        ExitReason::Syntax.exit();
                    },
                    b => {
                        eprintln!("esimpl: error: stanza {}: expected data command, \
                                   saw invalid byte {b:#02x}", stanzas.len());
                        ExitReason::Syntax.exit();
                    },
                }
                (data, eb) = parse_unary(&mut br, "stanza data", stanzas.len());
                if Semideque(semideque) == semideque_count {
                    ending_byte = eb;

                    // An output instruction could appear here.
                    while ending_byte == 6 || ending_byte == 7 {
                        data_line.push(Stanza(if ending_byte == 6 { 0 } else { 1 }));
                        ending_byte = br.next().transpose().exit_on_error().unwrap_or(0);
                    }
                    if !data_line.is_empty() {
                        push_data_command(DataCommandType::Output,
                                          mem::take(&mut data_line).into_boxed_slice(), &mut sb);
                    }

                    break;
                }
            }
        }
    }

    if !sb.current_data_commands.is_empty() {
        eprintln!("esimpl: error: the last stanza is missing a control instruction");
        ExitReason::Syntax.exit();
    }
    if sb.new_table {
        eprintln!("esimpl: error: empty table at end of input");
    }
    if stanzas.is_empty() {
        eprintln!("esimpl: error: the program requires at least one stanza");
        ExitReason::Syntax.exit();
    }

    Program(stanzas.into_boxed_slice())
}




/// Pretty-prints a program in the text format.
fn pretty_print_text(stream: &mut dyn io::Write, p: &Program) -> io::Result<()> {
    let mut bw = io::BufWriter::new(stream);

    for (num, commands) in p.0.iter().enumerate() {
        if commands.new_table {
            match commands.table_linked_to {
                None => panic!("table with no table link"),
                Some(TableLink::IoLinked) => write!(bw, "iotable\n\n")?,
                Some(TableLink::SemidequeLinked(Semideque(sdq))) => write!(bw, "{sdq} table\n\n")?,
            }
        }

        writeln!(bw, "# stanza {num}")?;
        for dc in commands.data_commands.iter() {
            match dc.cmd_type {
                DataCommandType::Push(Semideque(sdq)) => write!(bw, "{sdq} push")?,
                DataCommandType::PushBack(Semideque(sdq)) => write!(bw, "{sdq} pushback")?,
                DataCommandType::Output => write!(bw, "output")?,
            }
            for Stanza(pv) in dc.pushed_values.iter() {
                write!(bw, " {pv}")?;
            }
            writeln!(bw)?;
        }

        match commands.control_command {
            ControlCommand::Halt => write!(bw, "halt\n\n")?,
            ControlCommand::InputGoto(Stanza(stanza)) => write!(bw, "input-goto {stanza}\n\n")?,
            ControlCommand::PopGoto { pop_from: Semideque(pop_from), add: Stanza(add) } =>
                write!(bw, "{pop_from} pop-goto {add}\n\n")?,
            ControlCommand::Goto { linked_to: Semideque(linked_to), stanza: Stanza(stanza) } =>
                write!(bw, "{linked_to} goto {stanza}\n\n")?
        }
    }

    bw.flush()
}

/// Pretty-prints a program in the binary format.
fn pretty_print_binary(stream: &mut dyn io::Write, p: &Program) -> io::Result<()> {
    let mut bw = io::BufWriter::new(stream);
    let mut i = p.0.iter();

    fn dct_sort_key(dct: &DataCommand) -> i128 {
        use DataCommandType::*;
        match dct.cmd_type {
             Push(Semideque(sdq)) => (sdq as i128) * 2,
             PushBack(Semideque(sdq)) => (sdq as i128) * 2 + 1,
             Output => i128::MAX,
        }
    }
    fn dct_semideque(dct: &DataCommand) -> Option<(Semideque, bool)> {
        use DataCommandType::*;
        match dct.cmd_type {
            Push(sdq) => Some((sdq, false)),
            PushBack(sdq) => Some((sdq, true)),
            Output => None,
        }
    }

    let semideque_count;
    {
        let stanza0 = i.next().expect("program had no stanza 0");
        let ControlCommand::Goto { linked_to, stanza } = stanza0.control_command else {
            panic!("stanza 0 had no goto command")
        };
        let mut dc_copy = stanza0.data_commands.clone();
        dc_copy.sort_unstable_by_key(dct_semideque);
        let mut current_sdq = 0;
        if linked_to == Semideque(0) {
            for _ in 0..stanza.0 { bw.write_all(&[0])?; }
            bw.write_all(&[1])?;
        }
        for dc in dc_copy.iter() {
            while current_sdq < dct_semideque(&dc).expect("stanza 0 had an output command").0.0 {
                bw.write_all(&[2])?;
                current_sdq += 1;
                if Semideque(current_sdq) == linked_to {
                    for _ in 0..stanza.0 { bw.write_all(&[0])?; }
                    bw.write_all(&[1])?;
                }
            }
            for pv in dc.pushed_values.iter() {
                for _ in 0..(*pv).0 { bw.write_all(&[0])?; }
                bw.write_all(&[1])?;
            }
        }

        bw.write_all(&[2])?;
        bw.write_all(&[0xD])?;
        for _ in 0..linked_to.0 { bw.write_all(&[3, 2])?; }
        bw.write_all(&[8])?;
        semideque_count = current_sdq + 1;
    }

    for commands in i {
        if commands.new_table { bw.write_all(&[0xA])?; }

        for s in 0..semideque_count {
            match commands.table_linked_to.expect("non-0 stanza is in no table") {
                TableLink::IoLinked => bw.write_all(&[5])?,
                TableLink::SemidequeLinked(Semideque(sdq)) => bw.write_all(
                    if s < sdq { &[5] } else { &[4] })?
            }
        }

        let (goto_sort_key, goto_target, goto_trailing_1) =
            match commands.control_command {
                ControlCommand::Halt => (-1, Stanza(0), false),
                ControlCommand::PopGoto { pop_from, add: table } =>
                    ((pop_from.0 as i128) * 2, table, false),
                ControlCommand::Goto { stanza, linked_to } =>
                    ((linked_to.0 as i128) * 2, stanza, true),
                ControlCommand::InputGoto(table) => (0, table, true)
            };

        // Each sort key has to output a data command, so it makes more
        // sense to loop over the sort keys than data commands. The commands
        // are sorted in the same order, and `dc_cursor` keeps track of which
        // commands are still to be output.
        let mut dc_copy = commands.data_commands.clone();
        dc_copy.sort_by_key(dct_sort_key);
        let mut dc_cursor = &dc_copy[..];
        let mut target_sort_key = 0_i128;
        while target_sort_key < (semideque_count as i128) * 2 {
            let on_push = (target_sort_key & 1) == 0;

            if target_sort_key == goto_sort_key {
                for _ in 0..goto_target.0 { bw.write_all(&[0])?; }
                if goto_trailing_1 { bw.write_all(&[1])?; }
            }

            while let Some(curcmd) = dc_cursor.first() {
                if target_sort_key == dct_sort_key(curcmd) {
                    for pv in curcmd.pushed_values.iter() {
                        for _ in 0..(*pv).0 { bw.write_all(&[0])?; }
                        bw.write_all(&[1])?;
                    }
                    dc_cursor = &dc_cursor[1..];
                } else {
                    break;
                }
            }
            if on_push {
                bw.write_all(&[3])?;
            } else {
                bw.write_all(&[2])?;
            }
            target_sort_key += 1;
        }

        // Process the output commands.
        while let Some((outcmd, tail)) = dc_cursor.split_first() {
            assert_eq!(outcmd.cmd_type, DataCommandType::Output,
                       "a non-output command referred to a nonexistent semideque");
            dc_cursor = tail;

            for pv in outcmd.pushed_values.iter() {
                match pv {
                    Stanza(0) => bw.write_all(&[6])?,
                    Stanza(1) => bw.write_all(&[7])?,
                    Stanza(v) => panic!("an output command output invalid value {v}"),
                }
            }
        }

        // Process the control command (except for a goto location, which has
        // been processed already).
        match commands.control_command {
            ControlCommand::PopGoto { pop_from: Semideque(sdq), .. } |
            ControlCommand::Goto { linked_to: Semideque(sdq), .. } => {
                bw.write_all(&[9])?;
                for _ in 0..sdq { bw.write_all(&[3, 2])?; }
                bw.write_all(&[8])?;
            },
            ControlCommand::Halt => bw.write_all(&[0xC])?,
            ControlCommand::InputGoto(_) => bw.write_all(&[0xB])?
        }
    }

    bw.write_all(&[0xE])?;

    bw.flush()
}



/// Runs an Esimpl program.
fn run_program(program: &Program, debug_on_halt: bool, stdin: &mut impl io::Read) {
    let mut semideques: Vec<VecDeque<Stanza>> = Vec::new();
    let mut curstanza = 0_i128;
    let mut oldstanza = 0_i128;
    let mut input_byte: Option<u8> = None;
    let mut output_byte: u8 = 0;
    let mut expected_link: Option<TableLink> = None;

    fn get_semideque(sdqs: &mut Vec<VecDeque<Stanza>>, sdq: Semideque, loc: i128) -> &mut VecDeque<Stanza> {
        if let Some(rv) = sdqs.get_mut(sdq.0) { return rv; }
        eprintln!("esimpl: error: stanza {loc} refers to nonexistent semideque {}", sdq.0);
        ExitReason::Syntax.exit();
    }

    loop {
        let Some(stanza_commands) = usize::try_from(curstanza).ok().and_then(|v| program.0.get(v)) else {
            eprintln!("esimpl: error: stanza {oldstanza} attempted to run nonexistent stanza {curstanza}");
            ExitReason::Runtime.exit();
        };
        if stanza_commands.table_linked_to != expected_link {
            eprintln!("esimpl: error: stanza {oldstanza} transferred control to stanza {curstanza}, \
                       whose table is not linked correctly to receive control from there");
            ExitReason::Runtime.exit();
        }
        oldstanza = curstanza;

        if curstanza == 0 {
            // stanza 0 allocates the semideques in addition to pushing onto them,
            // so the check for inboundedness of the semideque number is different
            // from the other cases
            for dc in stanza_commands.data_commands.iter() {
                let DataCommandType::Push(sdq) = dc.cmd_type else {
                    eprintln!("esimpl: error: stanza 0 contained a command other than \
                               'push' and 'pop-goto'");
                    ExitReason::Runtime.exit();
                };
                while semideques.len() <= sdq.0 { semideques.push(VecDeque::new()); }

                if !semideques[sdq.0].is_empty() {
                    eprintln!("esimpl: error: stanza 0 pushed to semideque {} twice", sdq.0);
                    ExitReason::Runtime.exit();
                }
                // `push` and `pushback` commands should do the same thing in this case
                // because the semideque is empty; implement it as `pushback` because
                // doing so is more efficient
                semideques[sdq.0].extend(dc.pushed_values.iter());
            }
        } else {
            for dc in stanza_commands.data_commands.iter() {
                match dc.cmd_type {
                    DataCommandType::Push(sdq) => {
                        let sd = get_semideque(&mut semideques, sdq, curstanza);
                        // the command is written in the order the
                        // data is popped, thus the order needs to be reversed
                        for pv in dc.pushed_values.iter().rev() {
                            sd.push_front(*pv);
                        }
                    },
                    DataCommandType::PushBack(sdq) =>
                        get_semideque(&mut semideques, sdq, curstanza).extend(dc.pushed_values.iter()),
                    DataCommandType::Output => {
                        for pv in dc.pushed_values.iter() {
                            match pv {
                                Stanza(0) => {
                                    let Some(nob) = output_byte.checked_add(1) else {
                                        eprintln!("esimpl: error: stanza {curstanza} attempted to \
                                                   output a byte with value 256 or higher");
                                        ExitReason::Runtime.exit();
                                    };
                                    output_byte = nob;
                                },
                                Stanza(1) => {
                                    io::stdout().write(&[output_byte]).exit_on_error();
                                    output_byte = 0;
                                },
                                _ => {
                                    eprintln!("esimpl: error: stanza {curstanza} 'output' command \
                                               has an argument other than 0 or 1");
                                    ExitReason::Runtime.exit();
                                }
                            }
                        }
                    },
                }
            }
        }
        match stanza_commands.control_command {
            ControlCommand::Halt => {
                if debug_on_halt {
                    eprintln!("Halted at stanza {curstanza}. Final state of semideques \
                               (printed in front-at-left order):");
                    for (num, semideque_contents) in semideques.iter().enumerate() {
                        eprint!("Semideque {num}:");
                        for Stanza(s) in semideque_contents.iter() { eprint!(" {s}"); }
                        eprintln!("");
                    }
                }
                return;
            },
            ControlCommand::InputGoto(s) => {
                loop {
                    match input_byte {
                        None => {
                            let mut inbuf = [0; 1];
                            let read_len = stdin.read(&mut inbuf).exit_on_error();
                            if read_len < 1 {
                                // We reached EOF.
                                curstanza = (s.0 as i128) + 2;
                                break;
                            }
                            input_byte = Some(inbuf[0]);
                        },
                        Some(0) => {
                            input_byte = None;
                            curstanza = (s.0 as i128) + 1;
                            break;
                        },
                        Some(ref mut ib) => {
                            *ib -= 1;
                            curstanza = s.0 as i128;
                            break;
                        }
                    }
                }
                expected_link = Some(TableLink::IoLinked);
            },
            ControlCommand::Goto { linked_to, stanza } => {
                curstanza = stanza.0 as i128;
                expected_link = Some(TableLink::SemidequeLinked(linked_to));
            },
            ControlCommand::PopGoto { pop_from, add } => {
                let Some(popped_data) = get_semideque(
                    &mut semideques, pop_from, curstanza).pop_front() else {
                    eprintln!("esimpl: error: stanza {curstanza} popped from empty \
                               semideque {}", pop_from.0);
                    ExitReason::Runtime.exit();
                };
                curstanza = (popped_data.0 as i128) + (add.0 as i128);
                expected_link = Some(TableLink::SemidequeLinked(pop_from));
            },
        }
        if curstanza == 0 {
            eprintln!("esimpl: error: stanza {curstanza} attempted to rerun stanza 0");
            ExitReason::Runtime.exit();
        }
    }
}


fn usage(stream: &mut dyn io::Write, exit_code: ExitReason) -> ! {
    write!(stream,
           "Usage: esimpl [-b|-t] [-Pb|-Pt] [-D*] [program.esi]\n\
            \n    \
            -b   Read the program in binary syntax\n    \
            -t   Read the program in text syntax\n    \
            -P   Instead of running the program, pretty-print it in the given syntax\n    \
            -Dh  Output the state of all semideques when the program halts\n\
            \n\
            The program to be run can be specified on the command line.\n\
            If it isn't, it will be read from standard input.\n")
        .unwrap_or_else(|_| ExitReason::Io.exit());
    exit_code.exit();
}

fn usage_error(reason: &str, bad_arg: &str) -> ! {
    eprint!("esimpl: error: {reason}: {bad_arg:?}\n\n");
    usage(&mut io::stderr(), ExitReason::Usage);
}

fn main() {
    let mut parse_args = true;
    let mut in_file_name = None;
    let mut esi_syntax = EsiSyntax::Text;
    let mut pretty_print = None;
    let mut debug_on_halt = false;

    for arg in env::args_os().skip(1) {
        match arg.to_str().and_then(|a| a.strip_prefix('-')) {
            Some(a) if parse_args => match a {
                "-" => parse_args = false,
                "-help" => usage(&mut io::stdout(), ExitReason::Success),
                "-version" => {
                    let version = env!("CARGO_PKG_VERSION");
                    println!("Esimpl version {version}");
                    return;
                },
                "b" => esi_syntax = EsiSyntax::Binary,
                "t" => esi_syntax = EsiSyntax::Text,
                "Pb" => pretty_print = Some(EsiSyntax::Binary),
                "Pt" => pretty_print = Some(EsiSyntax::Text),
                "Dh" => debug_on_halt = true,
                _ => usage_error("Unrecognised command-line option", a)
            },
            _ => {
                if in_file_name == None {
                    in_file_name = Some(arg);
                } else {
                    usage_error("More than one filename given on the command line",
                                &arg.to_string_lossy());
                }
            }
        }
    }

    let mut stdin = io::stdin();
    let mut stdin = io::BufReader::new(&mut stdin);

    let program = if let Some(ifn) = in_file_name {
        let f = File::open(&ifn).unwrap_or_else(|e| {
            eprintln!("esimpl: error: could not open file {ifn:?}: {e}");
            ExitReason::OpenInput.exit();
        });
        parse_program(&mut io::BufReader::new(f), esi_syntax)
    } else {
        parse_program(&mut stdin, esi_syntax)
    };

    if let Some(pp) = pretty_print {
        match pp {
            EsiSyntax::Text => pretty_print_text(&mut io::stdout(), &program).exit_on_error(),
            EsiSyntax::Binary => pretty_print_binary(&mut io::stdout(), &program).exit_on_error(),
        };
        return;
    }

    run_program(&program, debug_on_halt, &mut stdin);
}
